package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experiment.*;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.ExperimentCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.rspec.*;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.dialogs.JFedBrowserFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.actions.ExperimentActionsFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.RDPAppUtil;
import be.iminds.ilabt.jfed.experimenter_gui.util.SSHTerminalUtil;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.gui_preferences.HLPreferenceKey;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.gui_preferences.JFedHLPreferences;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.*;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import javafx.application.Platform;
import javafx.scene.control.*;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.HBox;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 11/18/13
 * Time: 5:49 PM
 */
public class SliceExperimentCanvas extends ExperimentCanvas implements ExperimentChangeListener {
    private static final Logger LOG = LoggerFactory.getLogger(SliceExperimentCanvas.class);

    @Nonnull
    private final ModelRspecToCanvasBinding binding;

    @Nonnull
    private final ContextMenu linkContextMenu;
    @Nonnull
    private final ContextMenu addressPoolContextMenu;

    @Nonnull
    private final MenuItem shareLanItem;

    @Nonnull
    private final ExperimentController experimentController;
    @Nonnull
    private final Experiment experiment;
    @Nonnull
    private final ExperimentActionsFactory experimentActionsFactory;
    @Nonnull
    private final GeniUserProvider geniUserProvider;
    @Nonnull
    private final TestbedInfoSource testbedInfoSource;
    @Nonnull
    private final AuthorityFinder authorityFinder;
    @Nonnull
    private final HighLevelTaskFactory hltf;
    @Nonnull
    private final NodePropertiesDialogFactory nodePropertiesDialogFactory;
    @Nonnull
    private final RDPAppUtil rdpAppUtil;
    @Nonnull
    private final SSHTerminalUtil sshTerminalUtil;
    @Nonnull
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    @Nonnull
    private final JFedGuiConfig guiConfig;
    @Nonnull
    private final BrowserUtil browserUtil;
    @Nonnull
    private final JFedBrowserFactory jFedBrowserFactory;

    private final boolean isDemo = new JFedHLPreferences().getBoolean(HLPreferenceKey.PREF_IS_DEMO, false);


    public SliceExperimentCanvas(@Nonnull ExperimentController experimentController,
                                 @Nonnull FXModelRspec model,
                                 @Nonnull JFedGuiConfig config,
                                 @Nonnull CanvasNodeFactory canvasNodeFactory,
                                 @Nonnull ExperimentActionsFactory experimentActionsFactory,
                                 @Nonnull GeniUserProvider geniUserProvider,
                                 @Nonnull TestbedInfoSource testbedInfoSource,
                                 @Nonnull AuthorityFinder authorityFinder,
                                 @Nonnull HighLevelTaskFactory hltf,
                                 @Nonnull NodePropertiesDialogFactory nodePropertiesDialogFactory,
                                 @Nonnull SSHTerminalUtil sshTerminalUtil,
                                 @Nonnull RDPAppUtil rdpAppUtil,
                                 @Nonnull AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                 @Nonnull JFedGuiConfig guiConfig,
                                 @Nonnull BrowserUtil browserUtil,
                                 @Nonnull JFedBrowserFactory jFedBrowserFactory) {
        this.browserUtil = browserUtil;
        this.jFedBrowserFactory = jFedBrowserFactory;
        this.binding = new ModelRspecToCanvasBinding(model, this, config, canvasNodeFactory);

        this.experimentController = experimentController;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.experiment = experimentController.getExperiment();

        this.experimentActionsFactory = experimentActionsFactory;
        this.geniUserProvider = geniUserProvider;
        this.testbedInfoSource = testbedInfoSource;
        this.authorityFinder = authorityFinder;
        this.hltf = hltf;
        this.nodePropertiesDialogFactory = nodePropertiesDialogFactory;
        this.sshTerminalUtil = sshTerminalUtil;
        this.rdpAppUtil = rdpAppUtil;
        this.guiConfig = guiConfig;

        //change background-color of
        setCanvasStyle(CanvasStyle.FIXED);

        //initialize GUI

        /*
        Create node context menu
         */

        //contextMenu for links
        this.linkContextMenu = new ContextMenu();

        shareLanItem = new MenuItem("Share/Unshare Lan", GlyphUtils.createGlyph(FontAwesome.Glyph.SHARE_ALT));
        shareLanItem.setOnAction(actionEvent -> {
            RspecCanvasLink rspecCanvasLink = getSelectionProvider().getSelectedObject(RspecCanvasLink.class);

            if (rspecCanvasLink != null) {
                experimentActionsFactory.createShareLanAction(this.getScene().getWindow(), experimentController,
                        rspecCanvasLink.getRspecLink()).run();
            }
        });
        linkContextMenu.getItems().add(shareLanItem);


        MenuItem linkInfoMenuItem = new MenuItem("Show Link Info", GlyphUtils.createGlyph(FontAwesome.Glyph.INFO));
        linkInfoMenuItem.setOnAction(
                actionEvent -> {
                    RspecCanvasLink rspecCanvasLink = getSelectionProvider()
                            .getSelectedObject(RspecCanvasLink.class);

                    if (rspecCanvasLink != null) {
                        FXRspecLink manifestLink = tryGetManifestRspecLink(rspecCanvasLink);

                        if (manifestLink != null) {
                            new LinkPropertiesDialog(manifestLink).showDialog();
                        } else {
                            new Alert(Alert.AlertType.INFORMATION,
                                    "No link information available. (no manifest)")
                                    .show();
                        }
                    } else {
                        throw new NullPointerException("It should be impossible that rspecCanvasLink is null");
                    }
                });
        linkContextMenu.getItems().add(linkInfoMenuItem);


        experimentController.getExperiment().addExperimentChangeListener(this);
        updateCanvasStyle(experimentController.getExperiment().getExperimentState());

        //contextMenu for address pools
        this.addressPoolContextMenu = new ContextMenu();
        {
            MenuItem addressPoolInfoMenuItem =
                    new MenuItem("Show Address Pool Info", GlyphUtils.createGlyph(FontAwesome.Glyph.INFO));
            addressPoolInfoMenuItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedAddressPool() != null);
                AddressPoolCanvasNode addressPoolCanvasNode = getSelectionProvider().getSelectedAddressPool();

                FXAddressPool manifestAddressPool = tryGetManifestAddressPool(addressPoolCanvasNode);
                if (manifestAddressPool != null)
                    new AddressPoolPropertiesDialog(manifestAddressPool).show();
                else
                    new Alert(Alert.AlertType.INFORMATION, "No information available").show();
            });
            addressPoolContextMenu.getItems().add(addressPoolInfoMenuItem);
        }

        {
            MenuItem openBrowserMenuItem = new MenuItem("Open in browser", GlyphUtils.createGlyph(FontAwesome.Glyph.EXTERNAL_LINK));
            openBrowserMenuItem.setOnAction(event -> {

                assert (getSelectionProvider().getSelectedAddressPool() != null);
                AddressPoolCanvasNode addressPoolCanvasNode = getSelectionProvider().getSelectedAddressPool();

                FXAddressPool manifestAddressPool = tryGetManifestAddressPool(addressPoolCanvasNode);
                if (manifestAddressPool != null) {

                    Dialog<String> dialog = new Dialog<>();
                    dialog.setTitle("Open browser to node");
                    dialog.setHeaderText("Open browser to node");

                    HBox content = new HBox(5);
                    ComboBox<String> protocol = new ComboBox<String>();
                    protocol.setPrefWidth(110);
                    protocol.getItems().add("http://");
                    protocol.getItems().add("https://");

                    protocol.getSelectionModel().selectFirst();

                    ComboBox<String> hostname = new ComboBox<>();
                    hostname.setPrefWidth(400);

                    manifestAddressPool.getIPv4s().stream().map(IPv4::getAddress).forEach(hostname.getItems()::add);

                    hostname.getSelectionModel().selectFirst();

                    Spinner<Integer> portSpinner = new Spinner<>(0, 10240, 80);
                    portSpinner.setEditable(true);
                    portSpinner.setPrefWidth(100);

                    protocol.getSelectionModel().selectedItemProperty().addListener(
                            (prop, oldValue, newValue) -> {
                                if (Objects.equals(oldValue, newValue))
                                    return;
                                if (portSpinner.getValue() == 80 && Objects.equals(newValue, "https://"))
                                    portSpinner.getValueFactory().setValue(443);
                                if (portSpinner.getValue() == 443 && Objects.equals(newValue, "http://"))
                                    portSpinner.getValueFactory().setValue(80);
                            }
                    );

                    content.getChildren().addAll(new Label("Select the URL:"), protocol, hostname, new Label(":"), portSpinner);

                    dialog.getDialogPane().setContent(content);
                    dialog.getDialogPane().getButtonTypes().setAll(ButtonType.OK, ButtonType.CANCEL);

                    dialog.setResultConverter(param -> {
                        if (param == ButtonType.OK) {
                            return protocol.getSelectionModel().getSelectedItem() + hostname.getSelectionModel().getSelectedItem() + ":" + portSpinner.getValue().toString();
                        } else
                            return null;
                    });

                    dialog.showAndWait().ifPresent(browserUtil::openUrlInBrowser);
                }


            });
            addressPoolContextMenu.getItems().add(openBrowserMenuItem);
        }


    }

    private void showNodeContextMenu(@Nonnull RspecCanvasNode rspecCanvasNode, double posX, double posY) {
        if (experimentController.getExperiment().getSliceOpt().isEmpty()) {
            //should not happen, but we can handle it without crashing...
            LOG.error("showNodeContextMenu when experiment has NO slice. Will do nothing.");
            return;
        }
        @Nonnull FXRspecNode rspecNode = rspecCanvasNode.getRspecNode();
        @Nullable FXRspecNode manifestRspecNode = tryGetManifestRspecNode(rspecCanvasNode);

        ContextMenu nodeContextMenu = new ContextMenu();

        Slice slice = experimentController.getExperiment().getSliceOrNull();
        if(slice == null)
            return;

        @Nonnull List<BasicStringRspec.LoginService> loginServices = sshTerminalUtil.findAllLoginService(rspecNode, slice);
        @Nonnull Optional<BasicStringRspec.LoginService> bestLoginService = sshTerminalUtil.findBestLoginService(rspecNode, slice);
        boolean showSsh = bestLoginService.isPresent() || !loginServices.isEmpty();

        Set<String> possibleHostNames = new HashSet<>();
        if (showSsh) {
            loginServices.stream().map(BasicStringRspec.LoginService::getHostname).forEach(possibleHostNames::add);

            if (manifestRspecNode != null) {
                manifestRspecNode.getLoginServices().stream().map(LoginService::getHostname).forEach(possibleHostNames::add);
                manifestRspecNode.getInterfaces().stream().flatMap(iface -> iface.getIpAddresses().stream())
                        .map(RspecInterface.IpAddress::getAddress).forEach(possibleHostNames::add);
                if (manifestRspecNode.getHostName() != null)
                    possibleHostNames.add(manifestRspecNode.getHostName());
            }
        }

        if (showSsh && bestLoginService.isPresent()) {
            MenuItem sshMenuItem =
                    new MenuItem("Open SSH terminal", GlyphUtils.createGlyph(FontAwesome.Glyph.TERMINAL));
            //TODO: would be nice to show, but you can't put a tooltip here :/
            //String details = loginServiceToString(bestLoginService.get(), true, false);
            sshMenuItem.setOnAction(actionEvent -> { sshTerminalUtil.launchSSHTerminal(bestLoginService.get()); });
            nodeContextMenu.getItems().add(sshMenuItem);
        }

        assert this.geniUserProvider != null;
        assert this.geniUserProvider.getLoggedInGeniUser() != null;
        assert this.geniUserProvider.getLoggedInGeniUser().getUserUrn() != null;
        final String loggedInUsername = this.geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
        assert loggedInUsername != null;

        boolean showComplexOpenSsh = showSsh &&
                (!bestLoginService.isPresent() || loginServices.size() > 1 || !bestLoginService.get().getUsername().equalsIgnoreCase(loggedInUsername));

//        if (showSsh && loginServices.size() == 1 && bestLoginService.isPresent() && !bestLoginService.get().getUsername().equalsIgnoreCase(loggedInUsername)) {
//            MenuItem sshMenuItemForced =
//                    new MenuItem("Open SSH terminal with user \"" + loggedInUsername + "\"", GlyphUtils.createGlyph(FontAwesome.Glyph.TERMINAL));
//            sshMenuItemForced.setOnAction(actionEvent -> {
//                sshTerminalUtil.launchSSHTerminalForcedLocalUser(rspecNode, experimentController.getExperiment().getSliceOrNull());
//            });
//            nodeContextMenu.getItems().add(sshMenuItemForced);
//        }


        if (showSsh && !showComplexOpenSsh) {
            MenuItem sshCustomUserMenuItem =
                    new MenuItem("Open SSH terminal with custom user", GlyphUtils.createGlyph(FontAwesome.Glyph.TERMINAL));
            sshCustomUserMenuItem.setOnAction(event -> {
                String username1 =
                        geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();

                TextInputDialog dialog = new TextInputDialog(username1);
                dialog.setTitle("Open SSH terminal with custom user");
                dialog.setHeaderText("Open SSH terminal with custom user");
                dialog.setContentText("Please enter the custom username:");

                dialog.showAndWait()
                        .ifPresent(customUsername ->
                                sshTerminalUtil.launchSSHTerminal(rspecNode,
                                        experimentController.getExperiment().getSliceOrNull(),
                                        customUsername));

            });
            nodeContextMenu.getItems().add(sshCustomUserMenuItem);
        }

        if (showComplexOpenSsh) {
            List<String> usernames = loginServices.stream()
                    .map(BasicStringRspec.LoginService::getUsername)
                    .distinct()
                    .collect(Collectors.toList());
            if (!usernames.contains(loggedInUsername)) usernames.add(loggedInUsername);
            Map<String, List<BasicStringRspec.LoginService>> logins =
                    loginServices.stream().collect(Collectors.groupingBy(ls -> ls.toDescription(false, true)));

            Menu sshRootMenu =
                    new Menu("Open SSH terminal (host/user)", GlyphUtils.createGlyph(FontAwesome.Glyph.TERMINAL));
            nodeContextMenu.getItems().add(sshRootMenu);
//            MenuItem sshRootMenuTitle = new MenuItem("Host");
//            sshRootMenu.getItems().add(sshRootMenuTitle);
//            sshRootMenu.getItems().add(new SeparatorMenuItem());

            for (String login : logins.keySet()) {
                List<BasicStringRspec.LoginService> loginServiceList = logins.get(login);
                List<String> loginUsernames = loginServiceList.stream()
                        .map(BasicStringRspec.LoginService::getUsername)
                        .distinct()
                        .collect(Collectors.toList());

                Menu sshSubMenu = new Menu(login);
                sshRootMenu.getItems().add(sshSubMenu);
//                MenuItem sshSubMenuTitle = new MenuItem("User");
//                sshSubMenu.getItems().add(sshSubMenuTitle);
//                sshSubMenu.getItems().add(new SeparatorMenuItem());
                for (String username : usernames) {
                    MenuItem sshLogin = new MenuItem(username);
                    sshLogin.setOnAction(event -> {
                        sshTerminalUtil.launchSSHTerminal(loginServiceList.get(0), username);
                    });
                    if (!loginUsernames.contains(username)) {
                        sshLogin.setStyle("-fx-text-fill: grey;");
                    }
                    sshSubMenu.getItems().add(sshLogin);
                }

                MenuItem sshCustomUserMenuItem =
                        new MenuItem("Other user...");
                sshCustomUserMenuItem.setOnAction(event -> {
                    TextInputDialog dialog = new TextInputDialog(loggedInUsername);
                    dialog.setTitle("Open SSH terminal with custom user");
                    dialog.setHeaderText("Open SSH terminal with custom user");
                    dialog.setContentText("Please enter the custom username:");
                    dialog.showAndWait()
                            .ifPresent(customUsername ->
                                    sshTerminalUtil.launchSSHTerminal(loginServiceList.get(0), customUsername));

                });
                sshCustomUserMenuItem.setStyle("-fx-text-fill: grey;");
                sshSubMenu.getItems().add(new SeparatorMenuItem());
                sshSubMenu.getItems().add(sshCustomUserMenuItem);
            }
        }

        if (showSsh && manifestRspecNode != null) {
            MenuItem openBrowserMenuItem = new MenuItem("Open in browser", GlyphUtils.createGlyph(FontAwesome.Glyph.EXTERNAL_LINK));
            openBrowserMenuItem.setOnAction(event -> {
                OpenNodeBrowserDialog dialog =
                        new OpenNodeBrowserDialog(browserUtil, jFedBrowserFactory, possibleHostNames);
                dialog.doDialog();
            });
            nodeContextMenu.getItems().add(openBrowserMenuItem);
        }

        if (showSsh && manifestRspecNode != null) {
            MenuItem infoMenuItem = new MenuItem("Show Node Info", GlyphUtils.createGlyph(FontAwesome.Glyph.INFO));
            infoMenuItem.setOnAction(actionEvent -> {
                this.nodePropertiesDialogFactory.showSimplePropertiesDialog(
                        slice, manifestRspecNode,
                        bestLoginService.orElse(loginServices.get(0)),
                        loginServices
                );
            });
            nodeContextMenu.getItems().add(infoMenuItem);
        }


        boolean showCreateImage = false;
        boolean showReboot = false;
        boolean showReloadOs = false;
        boolean showOpenConsole = false;
        if (manifestRspecNode != null && manifestRspecNode.getComponentManagerId() != null) {
            //actually needed for GetVersion, but that is same case as CreateSliver
            Server selectedAuthority =
                    authorityFinder.findByUrn(manifestRspecNode.getComponentManagerId(), AuthorityFinder.Purpose.CREATE_SLIVER);

            if (selectedAuthority != null && !selectedAuthority.isEdgeVlan()) {
                try {
                    AggregateManagerWrapper amWrapper =
                            aggregateManagerWrapperFactory.getAggregateManagerWrapper(selectedAuthority);
                    showCreateImage =
                            amWrapper.areUserDiskImageFunctionsSupported() && manifestRspecNode.getSliverId() != null;
                    showReboot = amWrapper.isRestartSupported() && manifestRspecNode.getSliverId() != null;
                    showReloadOs =
                            amWrapper.isReloadOSSupported() && manifestRspecNode.getSliverId() != null;
                    showOpenConsole =
                            amWrapper.isOpenConsoleSupported() && manifestRspecNode.getSliverId() != null;
                } catch (Exception e) {
                    //most likely a NoAMException, see AutomaticAggregateManagerWrapper.choose() @ line 91
                    LOG.error("Error checking amWrapper for functionality (will be ignored)", e);
                }
            }
        }

        if (showCreateImage) {
            @Nonnull final MenuItem diskImageMenuItem = new MenuItem("Create Image", GlyphUtils.createGlyph(FontAwesome.Glyph.UPLOAD));
            diskImageMenuItem.setOnAction(actionEvent -> {
                experimentActionsFactory.createCreateDiskImageAction(getScene().getWindow(), experimentController, manifestRspecNode).run();
            });
            diskImageMenuItem.disableProperty().setValue(isDemo);

            nodeContextMenu.getItems().add(diskImageMenuItem);
        }


        if (showReboot) {
            @Nonnull final MenuItem rebootMenuItem = new MenuItem("Reboot", GlyphUtils.createGlyph(FontAwesome.Glyph.POWER_OFF));
            rebootMenuItem.setOnAction(actionEvent -> {
                Sliver selectedSliver = getSelectedSliver();
                if (selectedSliver != null) {
                    experimentActionsFactory.createRebootAction(getScene().getWindow(), experimentController, selectedSliver).run();
                } else {
                    LOG.error("Could not retrieve selected sliver sliver to reboot");
                }
            });
            rebootMenuItem.disableProperty().setValue(isDemo || guiConfig.isExperimentActionDisabled());
            nodeContextMenu.getItems().add(rebootMenuItem);
        }


        if (showReloadOs) {
            @Nonnull final MenuItem reloadMenuItem = new MenuItem("Reload OS", GlyphUtils.createGlyph(FontAwesome.Glyph.POWER_OFF));
            reloadMenuItem.setOnAction(actionEvent -> {
                Sliver selectedSliver = getSelectedSliver();
                if (selectedSliver != null) {
                    experimentActionsFactory.createReloadOSAction(getScene().getWindow(), experimentController, selectedSliver).run();
                } else {
                    LOG.error("Could not retrieve selected sliver sliver to reload os");
                }

            });
            reloadMenuItem.disableProperty().setValue(isDemo || guiConfig.isExperimentActionDisabled());
            nodeContextMenu.getItems().add(reloadMenuItem);
        }


        if (showOpenConsole) {
            @Nonnull final MenuItem openConsoleMenuItem = new MenuItem("Open Boot Console", GlyphUtils.createGlyph(FontAwesome.Glyph.TERMINAL));
            openConsoleMenuItem.setOnAction(actionEvent -> {
                Sliver selectedSliver = getSelectedSliver();
                if (selectedSliver != null) {
                    experimentActionsFactory.createOpenConsoleAction(getScene().getWindow(), experimentController, selectedSliver).run();
                } else {
                    LOG.error("Could not retrieve selected sliver sliver to open boot console");
                }
            });
            nodeContextMenu.getItems().add(openConsoleMenuItem);
        }

        if (showSsh) {
            @Nonnull final MenuItem uploadMenuItem = new MenuItem("Transfer files", GlyphUtils.createGlyph(FontAwesome.Glyph.UPLOAD));
            uploadMenuItem.setOnAction(actionEvent -> {
                experimentActionsFactory.createSCPUploadAction(getScene().getWindow(), experimentController, manifestRspecNode).run();
            });
            nodeContextMenu.getItems().add(uploadMenuItem);
        }

        nodeContextMenu.show(rspecCanvasNode, posX, posY);
    }


    /**
     * Tries to get the corresponding RspecNode from the manifest. Otherwise returns the one from the request
     *
     * @param rspecCanvasNode -
     * @return -
     */
    @Nullable
    private FXRspecNode tryGetManifestRspecNode(@Nonnull RspecCanvasNode rspecCanvasNode) {
        FXRspecNode rspecNode = rspecCanvasNode.getRspecNode();

        //try to fetch node from manifest rspec
        FXModelRspec manifestRspec = getFXManifestRspec();
        if (manifestRspec != null) {
            FXRspecNode manifestRspecNode = manifestRspec.getNodeByUniqueId(rspecNode.getUniqueId());
            if (manifestRspecNode == null && rspecNode.getClientId() != null) {
                //fallback to client-id

                manifestRspecNode = manifestRspec.getNodeByClientId(rspecNode.getClientId());
            }

            if (manifestRspecNode != null) {
                rspecNode = manifestRspecNode;
            }
        }
        return rspecNode;
    }

    /**
     * Tries to get the corresponding RspecNode from the manifest. Otherwise returns the one from the request
     *
     * @param rspecCanvasLink -
     * @return -
     */
    @Nullable
    private FXRspecLink tryGetManifestRspecLink(@Nonnull RspecCanvasLink rspecCanvasLink) {
        FXRspecLink rspecLink = rspecCanvasLink.getRspecLink();

        //try to fetch node from manifest rspec
        FXModelRspec manifestRspec = getFXManifestRspec();
        if (manifestRspec != null) {
            FXRspecLink manifestRspecLink = manifestRspec.getLinkByUniqueId(rspecLink.getUniqueId());
            if (manifestRspecLink == null && rspecLink.getClientId() != null) {
                //fallback to client-id
                manifestRspecLink = manifestRspec.getLinkByClientId(rspecLink.getClientId());
            }

            if (manifestRspecLink != null) {
                rspecLink = manifestRspecLink;
            }
        }
        return rspecLink;
    }

    @Nullable
    private FXAddressPool tryGetManifestAddressPool(@Nonnull AddressPoolCanvasNode addressPoolCanvasNode) {
        FXAddressPool canvasAddressPool = addressPoolCanvasNode.getAddressPool();

        FXModelRspec manifestRspec = getFXManifestRspec();
        if (manifestRspec != null) {
            return manifestRspec.getAddressPoolByClientId(canvasAddressPool.getClientId());
        } else
            return null;
    }

    @Nullable
    private FXModelRspec getFXManifestRspec() {
        if (experiment.getSliceOrNull() == null)
            return null;

        if (experiment.getSliceOrNull().getManifestRspec() == null)
            return null;

        FXModelRspec manifestRspec =
                experiment.getSliceOrNull().getManifestRspec().getModelRspec(ModelRspecType.FX);
        assert manifestRspec != null;

        return manifestRspec;
    }

    @Nonnull
    public static String statusToNodeStyleClass(@Nonnull NodeStatus status) {
        switch (status) {
            case READY_WITH_LOGINSERVICES:
                return "canvas-node-ready";
            case READY_WITHOUT_LOGINSERVICES:
                return "canvas-node-ready-nologin";
            case CHANGING:
                return "canvas-node-changing";
            case FAIL:
            case RESERVATION_FAIL:
                return "canvas-node-fail";
            case UNALLOCATED:
                return "canvas-node-unallocated";
            case UNINITIALISED:
                return "canvas-node-uninitialised";
            case UNKNOWN:
                return "canvas-node-unknown";
            default:
                throw new IllegalArgumentException();
        }
    }

    @Nonnull
    public static String statusToAddressPoolStyleClass(@Nonnull StatusDetails.SliverStatus status) {
        switch (status) {
            case READY:
                return "canvas-node-ready";
            case CHANGING:
                return "canvas-node-changing";
            case NOTREADY:
                return "canvas-node-notready";
            case FAIL:
                return "canvas-node-fail";
            case UNALLOCATED:
                return "canvas-node-unallocated";
            case UNINITIALISED:
                return "canvas-node-uninitialised";
            case UNKNOWN:
                return "canvas-node-unknown";
            default:
                throw new IllegalArgumentException();
        }
    }

    @Nonnull
    public static String statusToGTTStyleClass(@Nonnull StatusDetails.SliverStatus status) {
        switch (status) {
            case READY:
                return "canvas-node-ready";
            case CHANGING:
                return "canvas-node-changing";
            case NOTREADY:
                return "canvas-node-notready";
            case FAIL:
                return "canvas-node-fail";
            case UNALLOCATED:
                return "canvas-node-unallocated";
            case UNINITIALISED:
                return "canvas-node-uninitialised";
            case UNKNOWN:
                return "canvas-node-unknown";
            default:
                throw new IllegalArgumentException();
        }
    }

    @Nonnull
    public static String statusToInterfaceStyleClass(@Nonnull StatusDetails.SliverStatus status) {
        switch (status) {
            case READY:
                return "canvas-link-ready";
            case CHANGING:
                return "canvas-link-changing";
            case NOTREADY:
                return "canvas-link-notready";
            case FAIL:
                return "canvas-link-fail";
            case UNALLOCATED:
                return "canvas-link-unallocated";
            case UNINITIALISED:
                return "canvas-link-uninitialised";
            case UNKNOWN:
                return "canvas-link-unknown";
            default:
                throw new IllegalArgumentException();
        }
    }

    @Nonnull
    public static String statusToLinkCenterStyleClass(@Nullable StatusDetails.SliverStatus status) {
        if (status == null) {
            return "link-center-editable";
        }
        switch (status) {
            case READY:
                return "link-center-ready";
            case CHANGING:
                return "link-center-changing";
            case NOTREADY:
                return "link-center-notready";
            case FAIL:
                return "link-center-fail";
            case UNALLOCATED:
                return "link-center-unallocated";
            case UNINITIALISED:
                return "link-center-uninitialised";
            case UNKNOWN:
                return "link-center-unknown";
            default:
                throw new IllegalArgumentException();
        }
    }


    @Override
    public void canvasItemOnMouseClicked(MouseEvent mouseEvent) {
        super.canvasItemOnMouseClicked(mouseEvent);

        if (mouseEvent.getSource() instanceof RspecCanvasNode) {
            RspecCanvasNode node = (RspecCanvasNode) mouseEvent.getSource();

            getSelectionProvider().setSelectedCanvasNode(node);
            if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                //Open either RDP or SSH login
                if (rdpAppUtil.hasRdpService(node.getRspecNode(), experiment.getSliceOrNull())) {
                    rdpAppUtil.launchRdpApp(node.getRspecNode(), experiment.getSliceOrNull());
                } else {
                    sshTerminalUtil.launchSSHTerminal(node.getRspecNode(), experiment.getSliceOrNull());
                }
            } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                showNodeContextMenu(node, mouseEvent.getScreenX(), mouseEvent.getScreenY());
            }
        }

        if (mouseEvent.getSource() instanceof RspecCanvasLink.LinkCenter) {
            RspecCanvasLink link =
                    ((RspecCanvasLink.LinkCenter) mouseEvent.getSource()).getRspecCanvasLink();


            getSelectionProvider().setSelectedCanvasLink(link);
            assert (getSelectionProvider().getSelectedObject(RspecCanvasLink.class) != null);

            if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                //check if shareLanItem should be available
                RspecCanvasLink rspecCanvasLink = getSelectionProvider().getSelectedObject(RspecCanvasLink.class);

                boolean isStiching = rspecCanvasLink != null && rspecCanvasLink.getRspecLink().isStitched(testbedInfoSource);
                boolean isEmulab = false;
                boolean supportsSharedLan = false;

                if (link.getRspecLink().getComponentManagerUrns().size() == 1) {
                    //actually needed for GetVersion, but that is same case as CreateSliver
                    Server nodeAuth =
                            authorityFinder.findByUrn(link.getRspecLink().getComponentManagerUrns().get(0), AuthorityFinder.Purpose.CREATE_SLIVER);
                    if (nodeAuth != null) {
                        AggregateManagerWrapper amWrapper =
                                aggregateManagerWrapperFactory.getAggregateManagerWrapper(nodeAuth);
                        supportsSharedLan = amWrapper.isShareLanSupported();
                        if (nodeAuth.getServerType() != null)
                            isEmulab =
                                    nodeAuth.getServerType().equalsIgnoreCase("emulab") || nodeAuth.getServerType().equalsIgnoreCase("protogeni") || nodeAuth.getServerType().equalsIgnoreCase("instageni");
                    }
                }

                if (isEmulab && supportsSharedLan && !isStiching && !(isDemo || guiConfig.isExperimentActionDisabled()))
                    shareLanItem.setDisable(false);
                else
                    shareLanItem.setDisable(true);

                //show context menu
                if (getSelectionProvider().getSelectedObject(RspecCanvasLink.class) != null) {
                    RspecCanvasLink selectedRspecCanvasLink = getSelectionProvider().getSelectedObject(RspecCanvasLink.class);
                    Button linkCenter = selectedRspecCanvasLink == null ? null : selectedRspecCanvasLink.getLinkCenter();
                    linkContextMenu.show(linkCenter, mouseEvent.getScreenX(), mouseEvent.getScreenY());
                }
            }
        }

        if (mouseEvent.getSource() instanceof AddressPoolCanvasNode) {
            getSelectionProvider().setSelectedAddressPool((AddressPoolCanvasNode) mouseEvent.getSource());

            if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                addressPoolContextMenu.show(getSelectionProvider().getSelectedAddressPool(),
                        mouseEvent.getScreenX(), mouseEvent.getScreenY());
            } else if (mouseEvent.getClickCount() == 2) {
                AddressPoolCanvasNode addressPoolCanvasNode = (AddressPoolCanvasNode) mouseEvent.getSource();

                FXAddressPool manifestAddressPool = tryGetManifestAddressPool(addressPoolCanvasNode);
                if (manifestAddressPool != null)
                    new AddressPoolPropertiesDialog(manifestAddressPool).show();
                else
                    new Alert(Alert.AlertType.INFORMATION, "No information available").show();

            }

        }
    }


    @Override
    public void onExperimentStateChange(ExperimentState newExperimentState) {
        updateCanvasStyle(experiment.getExperimentState());

    }

    @Override
    public void onExperimentPartAdded(ExperimentPart experimentPart) {
        /* nothing to do */
    }

    private void updateCanvasStyle(ExperimentState newExperimentState) {
        CanvasStyle newStyle;

        switch (newExperimentState) {
            case EXPIRED:
            case UNKNOWN:
            case EMPTY:
                newStyle = CanvasStyle.EXPIRED;
                break;
            case FUTURE_RESERVATION:
                newStyle = CanvasStyle.FUTURE_RESERVATION;
                break;
            case TIMEOUT_WAITING:
            case FAILING:
            case FAILED:
                newStyle = CanvasStyle.FAILED;
                break;
            default:
                newStyle = CanvasStyle.FIXED;
        }

        Platform.runLater(() -> setCanvasStyle(newStyle));


    }

    public RspecGTSNode getRspecGTSNode(FXGeantTestbedType gtt) {
        return binding.getRspecGTSNode(gtt);
    }

    public AddressPoolCanvasNode getAddressPoolCanvasNode(FXAddressPool addressPool) {
        return binding.getAddressPoolCanvasNode(addressPool);
    }

    public RspecCanvasNode getRspecCanvasNode(FXRspecNode node) {
        return binding.getRspecCanvasNode(node);
    }

    public RspecCanvasLink getRspecCanvasLink(FXRspecLink link) {
        return binding.getRspecCanvasLink(link);
    }


    public enum NodeStatus {
        UNINITIALISED, /* nothing known about status*/
        READY_WITH_LOGINSERVICES,
        READY_WITHOUT_LOGINSERVICES,
        UNALLOCATED,   /* known not to exist */
        UNKNOWN,       /* known to exist, but no status known. probably it is changing. */
        CHANGING,
        FAIL,
        RESERVATION_FAIL;


        public static NodeStatus fromSliverStatus(@Nullable StatusDetails.SliverStatus sliverStatus, boolean loginServiceAvailable, boolean reservationFailed) {
            if (sliverStatus == null) {
                LOG.warn("NodeStatus.fromSliverStatus(null) called (will return UNKNOWN)");
                return UNKNOWN;
            }
            if (reservationFailed)
                return RESERVATION_FAIL;

            switch (sliverStatus) {
                case READY: {
                    if (loginServiceAvailable)
                        return READY_WITH_LOGINSERVICES;
                    else
                        return READY_WITHOUT_LOGINSERVICES;
                }
                case UNINITIALISED:
                    return UNINITIALISED;
                case UNALLOCATED:
                    return UNALLOCATED;
                case UNKNOWN:
                    return UNKNOWN;
                case FAIL:
                    return FAIL;
                case CHANGING:
                    return CHANGING;
                case NOTREADY:
                    return CHANGING; //NOTREADY is not the same as changing, but it should do.
                default: {
//                    return NodeStatus.valueOf(sliverStatus.name());
                    LOG.warn("NodeStatus.fromSliverStatus(" + sliverStatus.name() + ") called (unsupported, will return UNKNOWN)");
                    return UNKNOWN;
                }
            }
        }
    }


    @Nullable
    private FXRspecNode getSelectedManifestNode() {
        assert (getSelectionProvider().getSelectedCanvasNode() instanceof RspecCanvasNode);
        RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) getSelectionProvider().getSelectedCanvasNode();

        FXModelRspec manifestRspec = getFXManifestRspec();
        if (rspecCanvasNode != null && manifestRspec != null) {
            return manifestRspec.getNodeByUniqueId(rspecCanvasNode.getRspecNode().getUniqueId());
        } else {
            return null;
        }
    }

    private Sliver getSelectedSliver() {
        FXRspecNode selectedManifestNode = getSelectedManifestNode();

        if (selectedManifestNode != null) {
            if (selectedManifestNode.getSliverId() != null) {
                if (experimentController.getExperiment().getSliceOrNull() != null) {
                    return experimentController.getExperiment().getSliceOrNull().findSliver(selectedManifestNode.getSliverId());
                } else {
                    LOG.error("experimentController.getExperiment().getSlice() is null");
                    return null;
                }
            } else {
                LOG.error("In getSelectedSliver(): Manifest node {} does not have a sliver-Id!", selectedManifestNode.getUniqueId());
                return null;
            }
        } else {
            LOG.error("Could not retrieve the selected manifest node!");
            return null;
        }
    }

}
